
"""
Simple RAG service, performs query using graph RAG an LLM.
Input is query, output is response.
"""

import pulsar
from pulsar.schema import JsonSchema
import tempfile
import base64
import os
import argparse
import time

from ... schema import GraphRagQuery, GraphRagResponse
from ... log_level import LogLevel
from ... graph_rag import GraphRag

default_pulsar_host = os.getenv("PULSAR_HOST", 'pulsar://pulsar:6650')
default_input_queue = 'graph-rag-query'
default_output_queue = 'graph-rag-response'
default_subscriber = 'graph-rag'
default_graph_hosts = [ 'localhost' ]
default_vector_store = 'http://localhost:19530'

class Processor:

    def __init__(
            self,
            pulsar_host=default_pulsar_host,
            input_queue=default_input_queue,
            output_queue=default_output_queue,
            subscriber=default_subscriber,
            log_level=LogLevel.INFO,
            graph_hosts=default_graph_hosts,
            vector_store=default_vector_store,
            entity_limit=50,
            triple_limit=30,
            max_sg_size=3000,
    ):

        self.client = None

        self.client = pulsar.Client(
            pulsar_host,
            logger=pulsar.ConsoleLogger(log_level.to_pulsar())
        )

        self.consumer = self.client.subscribe(
            input_queue, subscriber,
            schema=JsonSchema(GraphRagQuery),
        )

        self.producer = self.client.create_producer(
            topic=output_queue,
            schema=JsonSchema(GraphRagResponse),
        )

        self.rag = GraphRag(
            pulsar_host=pulsar_host,
            graph_hosts=graph_hosts,
            vector_store=vector_store,
            verbose=True,
            entity_limit=entity_limit,
            triple_limit=triple_limit,
            max_sg_size=max_sg_size,
        )

    def run(self):

        while True:

            msg = self.consumer.receive()

            try:

                v = msg.value()

	        # Sender-produced ID

                id = msg.properties()["id"]

                print(f"Handling input {id}...", flush=True)

                response = self.rag.query(v.query)

                print("Send response...", flush=True)
                r = GraphRagResponse(response = response)
                self.producer.send(r, properties={"id": id})

                print("Done.", flush=True)

                # Acknowledge successful processing of the message
                self.consumer.acknowledge(msg)

            except Exception as e:

                print("Exception:", e, flush=True)

                # Message failed to be processed
                self.consumer.negative_acknowledge(msg)

    def __del__(self):

        if self.client:
            self.client.close()

def run():

    parser = argparse.ArgumentParser(
        prog='graph-rag',
        description=__doc__,
    )

    parser.add_argument(
        '-p', '--pulsar-host',
        default=default_pulsar_host,
        help=f'Pulsar host (default: {default_pulsar_host})',
    )

    parser.add_argument(
        '-i', '--input-queue',
        default=default_input_queue,
        help=f'Input queue (default: {default_input_queue})'
    )

    parser.add_argument(
        '-s', '--subscriber',
        default=default_subscriber,
        help=f'Queue subscriber name (default: {default_subscriber})'
    )

    parser.add_argument(
        '-o', '--output-queue',
        default=default_output_queue,
        help=f'Output queue (default: {default_output_queue})'
    )

    parser.add_argument(
        '-l', '--log-level',
        type=LogLevel,
        default=LogLevel.INFO,
        choices=list(LogLevel),
        help=f'Output queue (default: info)'
    )

    parser.add_argument(
        '-g', '--graph-hosts',
        default='cassandra',
        help=f'Graph hosts, comma separated (default: cassandra)'
    )

    parser.add_argument(
        '-v', '--vector-store',
        default='http://milvus:19530',
        help=f'Vector host (default: http://milvus:19530)'
    )

    parser.add_argument(
        '-e', '--entity-limit',
        type=int,
        default=50,
        help=f'Entity vector fetch limit (default: 50)'
    )

    parser.add_argument(
        '-t', '--triple-limit',
        type=int,
        default=30,
        help=f'Triple query limit, per query (default: 30)'
    )

    parser.add_argument(
        '-u', '--max-subgraph-size',
        type=int,
        default=3000,
        help=f'Max subgraph size (default: 3000)'
    )

    args = parser.parse_args()
    
    while True:

        try:

            p = Processor(
                pulsar_host=args.pulsar_host,
                input_queue=args.input_queue,
                output_queue=args.output_queue,
                subscriber=args.subscriber,
                log_level=args.log_level,
                graph_hosts=args.graph_hosts.split(","),
                vector_store=args.vector_store,
                entity_limit=args.entity_limit,
                triple_limit=args.triple_limit,
                max_sg_size=args.max_subgraph_size,
            )

            p.run()

        except Exception as e:

            print("Exception:", e, flush=True)
            print("Will retry...", flush=True)

        time.sleep(10)

